import torch
from torch.nn.functional import softplus
from torch import inf

def to_sigma(sigmatilde):       return torch.nn.functional.softplus(sigmatilde,beta=1,threshold=5)

def from_sigma(sigma):      return sigma + torch.log(-torch.expm1(-sigma))

def log_prob(dist="Normal", params=None, targets=None):
    if dist=="Normal":
        return -0.5*torch.log(2*torch.tensor(torch.pi)) - torch.log(params["sigma"]) - 0.5*((targets-params["mu"])/params["sigma"])**2
    else:
        raise ValueError("Unknown distribution.")

def kl_divergence(dist="Normal", params=None, prior_params=None):
    if dist=="Normal":
        var_ratio = (params["sigma"] / prior_params["sigma"]).pow(2)
        t1 = ((params["mu"] - prior_params["mu"]) / prior_params["sigma"]).pow(2)
        return 0.5 * (var_ratio + t1 - 1 - var_ratio.log())
    else:
        raise ValueError("Unknown distribution.")


import torch

def to_sigma(sigmatilde):
    return torch.log2(torch.pow(2,sigmatilde)+1)

def from_sigma(sigma):
    return torch.log2(torch.pow(2,sigma)-1)

def to_pi(pitilde):
    return torch.softmax(pitilde,dim=1)

def from_pi(pi):
    return torch.log(pi)-torch.log(pi).mean(dim=1)

def gaussian_kernel_exponent(x, mu, sigma):
    D = mu.shape[-1]
    x_mu = x.unsqueeze(1)-mu
    dist = torch.norm(x_mu,dim=-1)/sigma.squeeze()
    log_det = 2*D*torch.log(sigma).squeeze()
    return -0.5*(dist**2 + log_det + D*2*torch.tensor(torch.pi).log())

class IndexedDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        super(IndexedDataset,self).__init__()
        self.data = data

    def __getitem__(self, idx):
        return self.data[idx], idx

    def __len__(self):
        return len(self.data)


import torch
from ..utils.vae_utils import *

ACTIVATION = torch.nn.ELU()

class NNBlock(torch.nn.Module):
    def __init__(self, input_dim, output_dim, num_neurons=50, num_hidden_layers=2, **_):
        super(NNBlock, self).__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_neurons = num_neurons
        self.num_layers = num_hidden_layers

        self.input_layer = torch.nn.Linear(input_dim, num_neurons)
        self.middle_layers = torch.nn.ModuleList([torch.nn.Linear(num_neurons, num_neurons) for _ in range(num_hidden_layers)]) 
        self.output_layer = torch.nn.Linear(num_neurons, output_dim)

        # setup the non-linearity
        self.act = ACTIVATION

    def forward(self, x):
        h = x.view(-1, self.input_dim)
        h = self.act(self.input_layer(h))
        
        for layer in self.middle_layers:
            h = self.act(layer(h))

        h = self.output_layer(h)
        return h

    def _num_parameters(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

class ParameterizerNN(torch.nn.Module):
    def __init__(self, input_dim, output_dim, dist_params=["mu"], num_hidden_layers=2, num_neurons=50, **_):
        super(ParameterizerNN, self).__init__()
        self.dist_params = dist_params
        self.block_dict = torch.nn.ModuleDict()

        self.block_dict["input"] = NNBlock(input_dim, num_neurons, num_neurons=num_neurons, num_hidden_layers=num_hidden_layers)

        for param in dist_params:
            self.block_dict[param] = NNBlock(num_neurons, output_dim, num_neurons=num_neurons, num_hidden_layers=1)
        # setup the non-linearity
        self.act = ACTIVATION

    def forward(self, inputs):
        h = inputs.view(-1, self.block_dict["input"].input_dim)
        h = self.act(self.block_dict["input"](h))
        output_dict = {}
        for param in self.dist_params:
            output_dict[param] = self.block_dict[param](h)
        return output_dict
    
    def _num_parameters(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

class GaussianNN(torch.nn.Module):
    def __init__(self, input_dim, output_dim, learn_sigma=True, num_hidden_layers=2, num_neurons=50, **_):
        super(GaussianNN, self).__init__()

        self.learn_sigma = learn_sigma

        if learn_sigma: dist_params = ["mu", "sigma"]
        else: dist_params = ["mu"]

        self.parameterizer = ParameterizerNN(input_dim, output_dim, dist_params=dist_params, num_hidden_layers=num_hidden_layers, num_neurons=num_neurons)

    def _num_parameters(self):
        return self.parameterizer._num_parameters()
    
    def forward(self, inputs):
        param_dict = self.parameterizer(inputs)
        if self.learn_sigma: param_dict["sigma"] = to_sigma(param_dict["sigma"])
        else: param_dict["sigma"] = torch.ones_like(param_dict["mu"])
        return param_dict
    
    def rsample(self, param_dict=None, num_samples=1, **_):
        return param_dict["mu"] + param_dict["sigma"] * torch.randn((num_samples, *param_dict["mu"].shape))

    def sample(self, param_dict=None, num_samples=1, **_):
        return self.rsample(param_dict=param_dict, num_samples=num_samples)
    
    def log_likelihood(self, targets, param_dict=None):
        return log_prob("Normal", param_dict, targets)
    
    def kl_divergence(self, param_dict=None, prior_params={"mu":0.0, "sigma":1.0}):
        return kl_divergence("Normal", param_dict, prior_params)

def get_distribution_model(dist_type, **kwargs):
    if dist_type.lower() in ["gaussian", "gauss", "normal", "n", "g"]: return GaussianNN(**kwargs)
    else: raise NotImplementedError("Unknown distribution type: {}".format(dist_type))

def get_prior_params(dist_type, num_dims=1):
    if dist_type.lower() in ["gaussian", "gauss", "normal", "n", "g"]: return {"mu":torch.zeros(num_dims), "sigma":torch.ones(num_dims)}
    else: raise NotImplementedError("Unknown distribution type: {}".format(dist_type))


import torch
from ..utils.kde_utils import *

class VanillaKernelDensityEstimator(torch.nn.Module):
    """Vanilla Kernel Density Estimator"""
    def __init__(self, mu, sigma=None):
        """Initialize the model
        Args:
            mu (torch.Tensor): mu of the kernels
            sigma (torch.Tensor): sigma of the kernels"""
        super(VanillaKernelDensityEstimator,self).__init__()
        self.num_kernels, self.num_dims = mu.shape
        self.mu = mu.unsqueeze(0)
        if sigma is None: sigma = mu.std(0).mean()/self.num_kernels
        self.sigmatilde = from_sigma(torch.ones(1,self.num_kernels,1)*sigma)
        self.pitilde = from_pi(torch.ones(1,self.num_kernels)/self.num_kernels)

    def log_likelihood(self, x):
        """Compute the log likelihood of the model
        Args:
            x (torch.Tensor): input data
        Returns:
            log_likelihood (torch.Tensor): log likelihood of the model"""
        kernel_matrix_exponent = gaussian_kernel_exponent(x, self.mu, to_sigma(self.sigmatilde)) + torch.log(to_pi(self.pitilde))
        log_likelihood = torch.logsumexp(kernel_matrix_exponent,1)
        return log_likelihood
    
    def sample(self, num_samples=1):
        """Sample from the model
        Args:
            num_samples (int): number of samples to draw
        Returns:
            samples (torch.Tensor): samples from the model"""
        kernel_idx = torch.multinomial(to_pi(self.pitilde)[0], num_samples, replacement=True)
        mu_selected, sigma_selected = self.mu[0,kernel_idx,:], to_sigma(self.sigmatilde[0,kernel_idx])
        samples = sigma_selected * torch.randn(num_samples, self.num_dims) + mu_selected
        return samples
    
    def expectation_step(self, x, leave_one_out=False):
        """Expectation step of the EM algorithm
        Args:
            x (torch.Tensor): input data
            leave_one_out (bool): whether to leave one out
        Returns:
            responsibility_matrix (torch.Tensor): responsibility matrix"""
        kernel_matrix = torch.exp(gaussian_kernel_exponent(x, self.mu, to_sigma(self.sigmatilde)) + torch.log(to_pi(self.pitilde))) + 1e-32
        if leave_one_out: kernel_matrix = kernel_matrix*(1.0-torch.eye(self.num_kernels))
        responsibility_matrix = kernel_matrix/kernel_matrix.sum(1,keepdims=True)
        return responsibility_matrix
    
    def impute(self, x, num_steps=100):
        """Impute missing values
        Args:
            x (torch.Tensor): input data
            num_steps (int): number of steps to take
        Returns:
            x (torch.Tensor): imputed data"""
        miss_idx = x.isnan()
        x[miss_idx] = 0.0
        for _ in range(num_steps):
            kernel_idx = self.expectation_step(x, leave_one_out=False).argmax(1)
            sample = to_sigma(self.sigmatilde[0,kernel_idx]) * torch.randn(x.shape[0], self.num_dims) + self.mu[0,kernel_idx,:]
            x[miss_idx] = sample[miss_idx]
        return x
    
    def get_params(self):
        """Get the parameters of the model
        Returns:
            params (dict): parameters of the model"""
        return {'mu':self.mu, 'sigma':to_sigma(self.sigmatilde), 'pi':to_pi(self.pitilde)}
    
class AdaptiveKernelDensityEstimator(VanillaKernelDensityEstimator):
    def __init__(self, mu, sigma=0.1):
        super(AdaptiveKernelDensityEstimator,self).__init__(mu, sigma)
    
    def forward(self, x, leave_one_out=True, batch_idx=None):
        kernel_matrix_exponent = gaussian_kernel_exponent(x, self.mu, to_sigma(self.sigmatilde)) + torch.log(to_pi(self.pitilde))
        if leave_one_out:
            mask = 1.0-torch.eye(self.num_kernels)
            if batch_idx is not None: mask = mask[batch_idx]
            kernel_matrix_exponent = kernel_matrix_exponent*mask
            kernel_matrix_exponent[mask==0] = -1e32
        log_likelihood = torch.logsumexp(kernel_matrix_exponent,1)
        return log_likelihood
    
    def maximization_step(self, x, responsibility_matrix):
        x_mu = x.unsqueeze(1)-self.mu
        sigma = torch.sqrt(torch.sum(responsibility_matrix*(x_mu**2).mean(-1),0)/(responsibility_matrix.sum(0))).unsqueeze(-1)
        sigmatilde = from_sigma(sigma).unsqueeze(0)
        return sigmatilde, self.pitilde
    
    def _diff_parameterize(self):
        self.sigmatilde = torch.nn.Parameter(self.sigmatilde)

    def train(self, 
            x,
            leave_one_out=True,
            modified_em = False,
            learning_rate = 1e-2,
            bacth_size = 32,
            wait_convergence=True, 
            num_iterations=50, 
            objective_threshold=1e-4,
            verbose=False,
            verbose_freq=10):
        
        train_flag, itx, last_objective = True, 0, 0.0

        if not modified_em: 
            self._diff_parameterize()
            optimizer = torch.optim.Adam(self.parameters(), lr=learning_rate)
            dataloader = torch.utils.data.DataLoader(IndexedDataset(x), batch_size=bacth_size, shuffle=True, drop_last=True)
        
        while train_flag:
            if modified_em:
                self.sigmatilde, self.pitilde = self.maximization_step(x, self.expectation_step(x, leave_one_out=leave_one_out))
            else:
                for x_batch, idx in dataloader:
                    optimizer.zero_grad()
                    loss = -self(x_batch, leave_one_out=leave_one_out, batch_idx=idx).mean()
                    loss.backward()
                    optimizer.step()

            with torch.no_grad(): objective = self(x, leave_one_out=leave_one_out).mean()
            if objective.isnan(): raise ValueError('Objective Value is NaN!')

            itx+=1
            if verbose and itx%verbose_freq==0: print('Iteration: %d, Objective Value: %.5f'%(itx, objective))
    
            if wait_convergence:
                delta_objective = torch.abs(objective-last_objective)
                last_objective = objective
                if delta_objective<objective_threshold: train_flag=False
            else:
                if itx>=num_iterations: train_flag=False

class PiKernelDensityEstimator(AdaptiveKernelDensityEstimator):
    def __init__(self, mu, sigma=0.1):
        super(PiKernelDensityEstimator,self).__init__(mu, sigma)
    
    def _diff_parameterize(self):
        super()._diff_parameterize()
        self.pitilde = torch.nn.Parameter(self.pitilde)
    
    def maximization_step(self, x, responsibility_matrix):
        return super().maximization_step(x, responsibility_matrix)[0], from_pi(responsibility_matrix.sum(0,keepdim=True)/responsibility_matrix.sum())




import torch
from scipy.stats import ks_2samp as KS
from scipy.stats import cramervonmises_2samp as CVM

def MMDTest(base_samples, test_samples, alphas=[0.5, 1.0, 2.0, 5.0, 10.0]):
    
    n_base, n_test = base_samples.size(0), test_samples.size(0)
    
    a00 = 1.0/(n_base*(n_base - 1))
    a11 = 1.0/(n_test*(n_test - 1))
    a01 = -1.0/(n_base*n_test)

    sample_12 = torch.cat((base_samples, test_samples), 0)
    distances = torch.norm((sample_12.unsqueeze(1)-sample_12),dim=-1)

    kernels = None
    for alpha in alphas:
        kernels_a = torch.exp(-alpha*distances**2)
        if kernels is None: kernels = kernels_a
        else: kernels = kernels + kernels_a

    k_1 = kernels[:n_base, :n_base]
    k_2 = kernels[n_base:, n_base:]
    k_12 = kernels[:n_base, n_base:]

    score = (2*a01*k_12.sum() + 
            a00*(k_1.sum()-torch.trace(k_1)) + 
            a11*(k_2.sum()-torch.trace(k_2)))
    
    return score

def EnergyTest(base_samples, test_samples):

    n_base, n_test = base_samples.size(0), test_samples.size(0)

    a00 = -1.0/(n_base**2)
    a11 = -1.0/(n_test*2)
    a01 = 1.0/(n_base*n_test)

    sample_12 = torch.cat((base_samples, test_samples), 0)
    distances = torch.norm((sample_12.unsqueeze(1)-sample_12),dim=-1)

    d_1 = distances[:n_base, :n_base].sum()
    d_2 = distances[-n_test:, -n_test:].sum()
    d_12 = distances[:n_base, -n_test:].sum()

    score = 2*a01*d_12 + a00*d_1 + a11*d_2

    return score

def sample_comparison(model_samples, train_samples, test_samples, test="mmd", subsample_ratio=0.4, mc_runs=1000):
    assert model_samples.shape[0]==train_samples.shape[0]

    if test=="mmd": test_func = MMDTest
    elif test=="energy": test_func = EnergyTest
    else: raise NotImplementedError("Test not implemented.")

    num_samples = test_samples.shape[0]
    n = int(num_samples*subsample_ratio)

    base_scores, model_scores = torch.zeros(mc_runs), torch.zeros(mc_runs)

    for i in range(mc_runs):
        test_subsamples = test_samples[torch.randperm(num_samples)[:n]]
        base_scores[i] = test_func(test_subsamples,train_samples[torch.randperm(train_samples.shape[0])[:n]])
        model_scores[i] = test_func(test_subsamples,model_samples[torch.randperm(model_samples.shape[0])[:n]])
    return model_scores, base_scores

def model_comparison(model_scores, base_scores, test="ks"):
    if test=="ks": test_func = lambda x,y: KS(x,y).statistic
    elif test=="cvm": test_func = lambda x,y: CVM(x,y).statistic
    elif test=="mean": test_func = lambda x,y: y.mean()-x.mean()
    else: raise NotImplementedError("Test not implemented.")
    return test_func(base_scores, model_scores)



import torch
from .submodels import *
from ..utils.vae_utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
torch.autograd.set_detect_anomaly(True)

class VAE(torch.nn.Module):
    def __init__(self, 
                input_dim=None, 
                cond_dim=0,
                latent_dim=3,
                num_neurons=20,
                num_hidden_layers=2,
                learn_decoder_sigma=True,
                ):
        super(VAE, self).__init__()
        kwargs = locals()
        for key in ["self","__class__","input_dim","cond_dim"]: kwargs.pop(key)

        self.kwargs = kwargs
        for key in kwargs:
            setattr(self,key,kwargs[key])

        self.input_dim = input_dim
        self.cond_dim = cond_dim

        self.encoder = get_distribution_model("normal", input_dim=self.input_dim+self.cond_dim, output_dim=self.latent_dim, learn_sigma=True, **kwargs)
        self.decoder = get_distribution_model("normal", input_dim=self.latent_dim+self.cond_dim, output_dim=self.input_dim, learn_sigma=self.learn_decoder_sigma, **kwargs)
        self.prior_params = get_prior_params("normal", self.latent_dim)

        self.num_parameters = self.encoder._num_parameters()+self.decoder._num_parameters()
    
    def forward(self, inputs, conditions=None, mc_samples=1):
        if self.cond_dim==0: conditions = torch.Tensor(inputs.shape[0], 0)
        elif conditions==None: raise ValueError("Conditions must be provided.")

        posterior_params_dict = self.encoder(torch.cat((inputs,conditions),dim=1))
        z = self.encoder.rsample(posterior_params_dict, num_samples=mc_samples, **self.kwargs)

        likelihood_params_dict = self.decoder(torch.cat((z,conditions.unsqueeze(0).repeat_interleave(mc_samples,dim=0)),dim=2))
        for param in likelihood_params_dict:
            likelihood_params_dict[param] = likelihood_params_dict[param].view(mc_samples,inputs.shape[0],self.input_dim)
        
        return {"params":likelihood_params_dict}, {"params":posterior_params_dict, "samples": z}

    def sample(self, num_samples_prior=1, num_samples_likelihood=1, conditions=None):
        if self.cond_dim==0: conditions = torch.Tensor([])
        elif conditions==None: raise ValueError("Conditions must be provided.")
        conditions = conditions.unsqueeze(0).unsqueeze(0).repeat_interleave(num_samples_prior,dim=0)

        with torch.no_grad():
            z = self.encoder.sample(param_dict=self.prior_params, num_samples=num_samples_prior).unsqueeze(1)
            param_dict = self.decoder(torch.cat((z,conditions),dim=2))
            samples = self.decoder.sample(param_dict, num_samples=num_samples_likelihood)
            return {"params":param_dict, "samples": samples}
    
    def impute(self, x, conditions=None, num_steps=100, use_mean=False):
        miss_idx = x.isnan()
        x[miss_idx] = 0.0
        for _ in range(num_steps):
            x_dict, z_dict = self.reconstruct(x, conditions=conditions, mc_samples=1)
            if use_mean: x[miss_idx] = x_dict["params"]["mu"].squeeze()[miss_idx]
            else: x[miss_idx] = x_dict["samples"].squeeze()[miss_idx]
        return x
    
    def reconstruct(self, inputs, conditions=None, mc_samples=1):
        with torch.no_grad():
            x_dict, z_dict = self.forward(inputs, conditions, mc_samples)
            x_dict["samples"] = self.decoder.sample(x_dict["params"], num_samples=1)
            return x_dict, z_dict
    
    def reconstruction_loglikelihood(self, x, likelihood_params):
        return self.decoder.log_likelihood(x, likelihood_params).sum(dim=2).mean(dim=0)
    
    def kl_divergence(self, posterior_params, prior_params=None):
        if prior_params is None: prior_params = self.prior_params
        return self.encoder.kl_divergence(posterior_params, prior_params=self.prior_params).sum(dim=1)
    
    def loss(self, x, likelihood_params, posterior_params, prior_params=None, beta=1.0):
        rll = self.reconstruction_loglikelihood(x, likelihood_params).mean(dim=0)
        kl = self.kl_divergence(posterior_params,prior_params=prior_params).mean(dim=0)
        loss = -(rll-beta*kl)
        return {"loss":loss, "elbo": rll-kl, "rll": rll, "kl": kl}
    
    def train(self, 
            x,
            conditions=None,
            beta=1.0,
            mc_samples=1,
            learning_rate=1e-3,
            epochs=500,
            verbose_freq=100,
            batch_size=32,):
        
        if self.cond_dim==0: conditions = torch.Tensor(x.shape[0], 0)
        elif conditions==None: raise ValueError("Conditions must be provided.")
        dataset = torch.utils.data.TensorDataset(x, conditions)

        optim = torch.optim.Adam([{'params':self.encoder.parameters()}, {'params':self.decoder.parameters()}], lr=learning_rate)
        dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, drop_last=True)
        total_itx = ((dataset.__len__())//batch_size + (dataloader.drop_last==False))*epochs

        epx, itx = 0, 0
        for _ in range(epochs):
            epx += 1

            for inputs, conditions in dataloader:
                itx += 1

                # region Forward and Backward Pass
                x_dict, z_dict = self.forward(inputs, conditions, mc_samples=mc_samples)
                optim.zero_grad()
                loss = self.loss(inputs, x_dict["params"], z_dict["params"], prior_params=self.prior_params, beta=beta)
                loss["loss"].backward()
                optim.step()
                #endregion
                
                # region Logging
                if itx%verbose_freq==0: 
                    print(f"Iteration: {itx}/{total_itx} -- ELBO={loss['elbo'].item():.2e} / RLL={loss['rll'].item():.2e} / KL={loss['kl'].item():.2e}")
                #endregion